from typing import Any, Type, cast

from pydantic import BaseModel

from snowman.relation.table import (
    GenericInsertColumnTypedDict,
    GenericUpdateColumnTypedDict,
    Table,
)


def full_table_name(table: Type[Table]) -> str:
    return f"{table.__databas_name__}.{table.__schema_name__}.{table.__table_name__}"


def table_column_names(table: Type[Table]) -> list[str]:
    if issubclass(table, BaseModel):
        return [column_name for column_name in table.model_fields.keys()]

    else:
        raise ValueError(f"Table {table} is not a Pydantic model")


def table_columns_dict(
    table: Table[GenericInsertColumnTypedDict, GenericUpdateColumnTypedDict]
    | GenericInsertColumnTypedDict
    | GenericUpdateColumnTypedDict,
) -> dict[str, Any]:
    if isinstance(table, BaseModel):
        return table.model_dump()

    elif isinstance(table, dict):
        return cast(dict, table)

    else:
        raise ValueError(f"Table {table} is not a Pydantic model")



from typing import TypeAlias

from snowflake.connector.cursor import SnowflakeCursor

try:
    import pandas as pd  # type: ignore[import]

    PandasDataFrame: TypeAlias = pd.DataFrame  # type: ignore
    USE_PANDAS = True

except ImportError:

    class DataFrame:
        pass

    PandasDataFrame: TypeAlias = DataFrame  # type: ignore
    USE_PANDAS = False


try:
    import turu.snowflake  # type: ignore[import]

    TuruSnowflakeCursor = turu.snowflake.Cursor
    USE_TURU = True

except ImportError:
    TuruSnowflakeCursor = SnowflakeCursor
    USE_TURU = False



from typing import Callable, ClassVar, Generic, LiteralString, Type, TypedDict, TypeVar


class InsertColumnTypedDict(TypedDict):
    pass


class UpdateColumnTypedDict(TypedDict, total=False):
    pass


GenericInsertColumnTypedDict = TypeVar(
    "GenericInsertColumnTypedDict",
    bound=InsertColumnTypedDict,
)

GenericUpdateColumnTypedDict = TypeVar(
    "GenericUpdateColumnTypedDict",
    bound=UpdateColumnTypedDict,
)


class Table(Generic[GenericInsertColumnTypedDict, GenericUpdateColumnTypedDict]):
    __databas_name__: ClassVar[str]
    __schema_name__: ClassVar[str]
    __table_name__: ClassVar[str]
    __insert_columns__: Type[GenericInsertColumnTypedDict] | None = None
    __update_columns__: Type[GenericUpdateColumnTypedDict] | None = None


GenericTable = TypeVar("GenericTable", bound=Table)


def table(
    database_name: LiteralString,
    schema_name: LiteralString,
    table_name: LiteralString,
    /,
) -> "Callable[[Type[GenericTable]], Type[GenericTable]]":
    def decorate(cls: "Type[GenericTable]") -> "Type[GenericTable]":
        cls.__databas_name__ = database_name
        cls.__schema_name__ = schema_name
        cls.__table_name__ = table_name

        return cls

    return decorate



"""
Snowman provides simple query builder for `insert` / `update` / `delete` / `truncate`.
"""

from typing import Type

from snowman.query._builder._delete_query_builder import DeleteQueryBuilder
from snowman.query._builder._insert_query_builder import InsertQueryBuilder
from snowman.query._builder._truncate_query_builder import TruncateQueryBuilder
from snowman.query._builder._update_query_builder import UpdateStatement
from snowman.relation.table import (
    GenericInsertColumnTypedDict,
    GenericUpdateColumnTypedDict,
    Table,
)

insert = InsertQueryBuilder()
"""
Insert query builder.
"""


def update(
    table: Type[Table[GenericInsertColumnTypedDict, GenericUpdateColumnTypedDict]],
):
    """
    Update query builder.
    """
    return UpdateStatement(
        table,
        _columns_type=table.__update_columns__,  # type: ignore
    )


delete = DeleteQueryBuilder()
"""
Delete query builder.
"""

truncate = TruncateQueryBuilder()
"""
Truncate query builder.
"""


__all__ = [
    "insert",
    "update",
    "delete",
    "truncate",
]



"""
# [Snowman](https://github.com/yassun7010/snowman-py)

Python model and query builder for [Snowflake](https://www.snowflake.com/).
"""

import importlib
import importlib.metadata

from . import datatype as datatype
from . import query as query
from .relation.table import Table as Table
from .relation.table import table as table

__version__ = importlib.metadata.version("snowman-py")



from typing import Any, Generic, Sequence, Type

from typing_extensions import override

from snowman.relation import full_table_name
from snowman.relation.table import GenericTable

from ._builder import QueryBuilder, QueryWithParams


class DeleteQueryBuilder:
    def from_(
        self, table: Type[GenericTable], /
    ) -> "DeleteFromStatement[GenericTable]":
        return DeleteFromStatement(table)


class DeleteFromStatement(Generic[GenericTable]):
    def __init__(self, table: Type[GenericTable]):
        self._table = table

    def where(
        self,
        condition: str,
        params: Sequence[Any] | None = None,
        /,
    ) -> "UpdateFromWhereQueryBuilder[GenericTable]":
        """
        Specify the condition of the where clause.

        Query parameters only support positional placeholders, so specify them with `%s`.

        e.g)
            `.where("id = %s AND name = %s", [1, "Alice"])`
        """
        return UpdateFromWhereQueryBuilder(self._table, condition, params or ())


class UpdateFromWhereQueryBuilder(Generic[GenericTable], QueryBuilder):
    def __init__(
        self,
        table: Type[GenericTable],
        where_condition: str,
        where_params: Sequence[Any],
    ):
        self._table = table
        self._where_condition = where_condition
        self._where_params = where_params

    @override
    def build(self) -> QueryWithParams:
        query = f"""
DELETE FROM
    {full_table_name(self._table)}
WHERE
    {self._where_condition}
""".strip()

        return QueryWithParams(query, tuple(self._where_params))



from abc import ABC, abstractmethod
from typing import Any, NamedTuple

from snowman.cursor import Cursor


class QueryWithParams(NamedTuple):
    query: str
    params: tuple[Any, ...]


class QueryBuilder(ABC):
    @abstractmethod
    def build(self) -> QueryWithParams: ...

    def execute(self, cursor: Cursor, /) -> None:
        query, params = self.build()

        if isinstance(params, tuple):
            cursor.executemany(query, params)
        else:
            cursor.execute(query, params)



from typing import Generic, Type

from typing_extensions import override

from snowman.relation import full_table_name
from snowman.relation.table import GenericTable

from ._builder import QueryBuilder, QueryWithParams


class TruncateQueryBuilder:
    @property
    def table(self) -> "TruncateTableStatement":
        return TruncateTableStatement()

    def __call__(
        self, table: Type[GenericTable], /
    ) -> "TruncateTableQueryBuilder[GenericTable]":
        return TruncateTableQueryBuilder(table, if_exists=False)

    @property
    def if_(self) -> "TruncateTableIfStatement":
        return TruncateTableIfStatement()


class TruncateTableStatement:
    def __call__(
        self, table: Type[GenericTable], /
    ) -> "TruncateTableQueryBuilder[GenericTable]":
        return TruncateTableQueryBuilder(table, if_exists=False)

    @property
    def if_(self) -> "TruncateTableIfStatement":
        return TruncateTableIfStatement()


class TruncateTableIfStatement:
    def exists(
        self, table: Type[GenericTable], /
    ) -> "TruncateTableQueryBuilder[GenericTable]":
        return TruncateTableQueryBuilder(table, if_exists=True)


class TruncateTableQueryBuilder(Generic[GenericTable], QueryBuilder):
    def __init__(self, table: Type[GenericTable], /, *, if_exists: bool):
        self._table = table
        self._if_exists = if_exists

    @override
    def build(self) -> QueryWithParams:
        query = f"TRUNCATE TABLE{ ' IF EXISTS' if self._if_exists else ''} {full_table_name(self._table)}"

        return QueryWithParams(query, ())



import itertools
from typing import Any, Generic, Sequence, Type, cast

from pydantic import BaseModel
from typing_extensions import override

from snowman.relation import full_table_name
from snowman.relation.table import GenericTable, GenericUpdateColumnTypedDict

from ._builder import QueryBuilder, QueryWithParams


class UpdateStatement(Generic[GenericTable, GenericUpdateColumnTypedDict]):
    def __init__(
        self,
        table: Type[GenericTable],
        /,
        *,
        _columns_type: Type[GenericUpdateColumnTypedDict] | None = None,
    ) -> None:
        self._table = table

    def set(
        self,
        fields: GenericTable | GenericUpdateColumnTypedDict,
    ) -> "UpdateSetQueryBuilder[GenericTable, GenericUpdateColumnTypedDict]":
        return UpdateSetQueryBuilder(self._table, fields)


class UpdateSetQueryBuilder(Generic[GenericTable, GenericUpdateColumnTypedDict]):
    def __init__(
        self,
        table: Type[GenericTable],
        columns: GenericTable | GenericUpdateColumnTypedDict,
    ):
        self._table = table
        self._columns = cast(
            dict,
            columns.model_dump(exclude_unset=True)
            if isinstance(columns, BaseModel)
            else columns,
        )

    def where(
        self,
        condition: str,
        params: Sequence[Any] | None = None,
        /,
    ) -> "UpdateSetWhereQueryBuidler":
        """
        Specify the condition of the where clause.

        Query parameters only support positional placeholders, so specify them with `%s`.

        e.g)
            `.where("id = %s AND name = %s", [1, "Alice"])`
        """
        return UpdateSetWhereQueryBuidler(
            self._table,
            self._columns,
            where_condition=condition,
            where_params=params or (),
        )


class UpdateSetWhereQueryBuidler(
    Generic[GenericTable, GenericUpdateColumnTypedDict], QueryBuilder
):
    def __init__(
        self,
        table: type[GenericTable],
        columns: GenericTable | GenericUpdateColumnTypedDict,
        *,
        where_condition: str,
        where_params: Sequence[Any],
    ):
        self._table = table
        self._columns = cast(
            dict,
            columns.model_dump(exclude_unset=True)
            if isinstance(columns, BaseModel)
            else columns,
        )
        self._where_condition = where_condition
        self._where_params = where_params

    @override
    def build(self) -> QueryWithParams:
        values = ",\n    ".join([f"{key} = %s" for key in self._columns.keys()])

        query = f"""
UPDATE
    {full_table_name(self._table)}
SET
    {values}
WHERE
    {self._where_condition}
""".strip()

        return QueryWithParams(
            query,
            tuple(itertools.chain(self._columns.values(), self._where_params)),
        )



from typing import Generic, Sequence, Type

from typing_extensions import override

import snowman._features
from snowman.cursor import Cursor, _get_snowflake_connection
from snowman.exception import snowmanNotDataFrameAvailableError
from snowman.relation import full_table_name, table_column_names, table_columns_dict
from snowman.relation.table import (
    GenericInsertColumnTypedDict,
    GenericTable,
    GenericUpdateColumnTypedDict,
    Table,
)

from ._builder import QueryBuilder, QueryWithParams


class InsertQueryBuilder:
    """
    Insert records into a table.

    >>> from snowman.query import insert
    >>> from your.database.schema import User
    >>>
    >>> query, _ = (
    ...     insert.into(
    ...         User
    ...     ).values(
    ...         {"id": 1, "name": "Alice"}
    ...     )
    ... ).build()
    >>>
    >>> print(query)
    INSERT INTO
        database.schema.users
    (
        id,
        name
    )
    VALUES (
        %s,
        %s
    )
    """

    @property
    def overwrite(self) -> "InsertOverwriteQueryBuilder":
        return InsertOverwriteQueryBuilder()

    def into(
        self,
        table: Type[Table[GenericInsertColumnTypedDict, GenericUpdateColumnTypedDict]],
        /,
    ):
        return InsertIntoQueryBuilder(
            table,
            _columns_type=table.__insert_columns__,  # type: ignore
        )


class InsertOverwriteQueryBuilder:
    def into(
        self,
        table: Type[Table[GenericInsertColumnTypedDict, GenericUpdateColumnTypedDict]],
        /,
    ):
        return InsertIntoQueryBuilder(
            table,
            overwtire=True,
            _columns_type=table.__insert_columns__,  # type: ignore
        )


class InsertIntoQueryBuilder(Generic[GenericTable, GenericInsertColumnTypedDict]):
    def __init__(
        self,
        table: Type[GenericTable],
        *,
        overwtire: bool = False,
        _columns_type: Type[GenericInsertColumnTypedDict] | None,
    ) -> None:
        self._table = table
        self._overwrite = overwtire

    def values(
        self,
        values: GenericTable
        | GenericInsertColumnTypedDict
        | Sequence[GenericTable | GenericInsertColumnTypedDict]
        | snowman._features.PandasDataFrame,
        /,
    ) -> "InsertIntoValuesQueryBuilder[GenericTable, GenericInsertColumnTypedDict]":
        if isinstance(values, snowman._features.PandasDataFrame):
            dataframe = values
            values = ()

        else:
            dataframe = None
            if not isinstance(values, Sequence):
                values = (values,)

        return InsertIntoValuesQueryBuilder(
            self._table,
            values,
            dataframe=dataframe,
            overwrite=self._overwrite,
        )


class InsertIntoValuesQueryBuilder(
    Generic[GenericTable, GenericInsertColumnTypedDict], QueryBuilder
):
    def __init__(
        self,
        table: type[GenericTable],
        values: Sequence[GenericTable | GenericInsertColumnTypedDict],
        /,
        *,
        dataframe: snowman._features.PandasDataFrame | None = None,
        overwrite: bool = False,
    ):
        self._table = table
        self._values = values
        self._dataframe = dataframe
        self._overwrite = overwrite

    @override
    def build(self) -> QueryWithParams:
        if self._dataframe is not None:
            raise snowmanNotDataFrameAvailableError()

        overwrite = "OVERWRITE " if self._overwrite else ""
        names = table_column_names(self._table)
        keys = ",\n    ".join(names)
        values = ",\n    ".join("%s" for _ in names)

        query = f"""
INSERT {overwrite}INTO
    {full_table_name(self._table)}
(
    {keys}
)
VALUES (
    {values}
)
""".strip()

        return QueryWithParams(
            query,
            tuple(table_columns_dict(self._values[0]).values())
            if len(self._values) == 1
            else tuple(
                tuple(table_columns_dict(value).values()) for value in self._values
            ),
        )

    @override
    def execute(self, cursor: Cursor, /) -> None:
        if self._dataframe is not None:
            from snowflake.connector.pandas_tools import write_pandas

            write_pandas(
                conn=_get_snowflake_connection(cursor),
                df=self._dataframe,
                table_name=self._table.__table_name__,
                database=self._table.__databas_name__,
                schema=self._table.__schema_name__,
            )
            return

        else:
            super().execute(cursor)
